"""API Views and ViewSets"""

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models.loading import get_models
from django.http import Http404
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from djes.apps import indexable_registry

import elasticsearch
from elasticsearch_dsl.query import Q
from elasticsearch_dsl.filter import F

from rest_framework import (
    filters,
    status,
    viewsets,
    routers
)
from firebase_token_generator import create_token

from rest_framework.decorators import detail_route, list_route
from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response

from bulbs.content.custom_search import custom_search_model
from bulbs.content.models import Content, Tag, LogEntry, FeatureType, ObfuscatedUrlInfo
from bulbs.content.serializers import (
    ContentSerializer, LogEntrySerializer, PolymorphicContentSerializer,
    TagSerializer, UserSerializer, FeatureTypeSerializer,
    ObfuscatedUrlInfoSerializer
)

from bulbs.contributions.serializers import ContributionSerializer
from bulbs.contributions.models import Contribution

from .mixins import UncachedResponse
from .permissions import CanEditContent, CanPublishContent


class ContentViewSet(UncachedResponse, viewsets.ModelViewSet):
    """
    uncached viewset for the `bulbs.content.Content` model
    """

    model = Content
    queryset = Content.objects.all()
    serializer_class = PolymorphicContentSerializer
    include_base_doctype = False
    paginate_by = 20
    filter_fields = ("search", "before", "after", "status", "feature_types", "published", "tags", "types")
    permission_classes = [IsAdminUser, CanEditContent]

    def get_serializer_class(self):
        """gets the class type of the serializer

        :return: `rest_framework.Serializer`
        """
        klass = None

        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        if lookup_url_kwarg in self.kwargs:
            # Looks like this is a detail...
            klass = self.get_object().__class__
        elif "doctype" in self.request.REQUEST:
            base = self.model.get_base_class()
            doctypes = indexable_registry.families[base]
            try:
                klass = doctypes[self.request.REQUEST["doctype"]]
            except KeyError:
                raise Http404

        if hasattr(klass, "get_serializer_class"):
            return klass.get_serializer_class()

        # TODO: fix deprecation warning here -- `get_serializer_class` is going away soon!
        return super(ContentViewSet, self).get_serializer_class()

    def post_save(self, obj, created=False):
        """indexes the object to ElasticSearch after any save function (POST/PUT)

        :param obj: instance of the saved object
        :param created: boolean expressing if object is newly created (`False` if updated)
        :return: `rest_framework.viewset.ModelViewSet.post_save`
        """
        from bulbs.content.tasks import index

        index.delay(obj.polymorphic_ctype_id, obj.pk)

        message = "Created" if created else "Saved"
        LogEntry.objects.log(self.request.user, obj, message)
        return super(ContentViewSet, self).post_save(obj, created=created)

    def list(self, request, *args, **kwargs):
        """Modified list view to driving listing from ES"""
        search_kwargs = {"published": False}

        for field_name in ("before", "after", "status", "published"):

            if field_name in self.request.QUERY_PARAMS:
                search_kwargs[field_name] = self.request.QUERY_PARAMS.get(field_name)

        for field_name in ("tags", "types", "feature_types"):

            if field_name in self.request.QUERY_PARAMS:
                search_kwargs[field_name] = self.request.QUERY_PARAMS.getlist(field_name)

        if "search" in self.request.QUERY_PARAMS:
            search_kwargs["query"] = self.request.QUERY_PARAMS.get("search")

        queryset = Content.search_objects.search(**search_kwargs)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    @detail_route(permission_classes=[CanPublishContent], methods=['post'])
    def publish(self, request, **kwargs):
        """sets the `published` value of the `Content`

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """
        content = self.get_object()

        if "published" in request.DATA:
            if not request.DATA["published"]:
                content.published = None
            else:
                publish_dt = parse_datetime(request.DATA["published"])
                if publish_dt:
                    publish_dt = publish_dt.astimezone(timezone.utc)
                else:
                    publish_dt = None
                content.published = publish_dt
        else:
            content.published = timezone.now()

        content.save()
        LogEntry.objects.log(request.user, content, content.get_status())
        return Response({"status": content.get_status(), "published": content.published})

    @detail_route(permission_classes=[CanPublishContent], methods=['post'])
    def trash(self, request, **kwargs):
        """destroys a `Content` instance and removes it from the ElasticSearch index

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        :raise: 404
        """
        content = self.get_object()

        content.indexed = False
        content.save()

        index = content.__class__.search_objects.mapping.index
        doc_type = content.__class__.search_objects.mapping.doc_type

        try:
            Content.search_objects.client.delete(
                index=index,
                doc_type=doc_type,
                id=content.id)
            LogEntry.objects.log(request.user, content, "Trashed")
            return Response({"status": "Trashed"})
        except elasticsearch.exceptions.NotFoundError:
            return Response(status=status.HTTP_404_NOT_FOUND)

    @detail_route(methods=["get"])
    def status(self, request, **kwargs):
        """This endpoint returns a status text, currently one of:
          - "Draft" (If no publish date is set, and no item exists in the editor queue)
          - "Waiting for Editor" (If no publish date is set, and an item exists in the editor queue)
          - "Published" (The published date is in the past)
          - "Scheduled" (The published date is set in the future)

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """
        content = self.get_object()
        return Response({"status": content.get_status()})

    @detail_route(methods=["post", "get"])
    def contributions(self, request, **kwargs):
        """gets or adds contributions

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """
        # Check if the contribution app is installed
        if Contribution not in get_models():
            return Response([])
        queryset = Contribution.objects.filter(content=self.get_object())
        if request.method == "POST":
            serializer = ContributionSerializer(
                queryset,
                data=request.DATA,
                many=True)
            if not serializer.is_valid():
                return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)
            serializer.save()
            return Response(serializer.data)
        else:
            serializer = ContributionSerializer(queryset, many=True)
            return Response(serializer.data)

    @detail_route(methods=["post"], permission_classes=[CanEditContent])
    def create_token(self, request, **kwargs):
        """Create a new obfuscated url info to use for accessing unpublished content.

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """

        data = {
            "content": self.get_object().id,
            "create_date": request.DATA["create_date"],
            "expire_date": request.DATA["expire_date"]
        }
        serializer = ObfuscatedUrlInfoSerializer(data=data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST, content_type="application/json")
        serializer.save()

        return Response(serializer.data, status=status.HTTP_200_OK, content_type="application/json")

    @detail_route(methods=["get"], permission_classes=[CanEditContent])
    def list_tokens(self, request, **kwargs):
        """List all tokens for this content instance.

        :param request: a WSGI request object
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """

        # no date checking is done here to make it more obvious if there's an issue with the
        # number of records. Date filtering will be done on the frontend.
        infos = [ObfuscatedUrlInfoSerializer(info).data
                 for info in ObfuscatedUrlInfo.objects.filter(content=self.get_object())]
        return Response(infos, status=status.HTTP_200_OK, content_type="application/json")


class TagViewSet(UncachedResponse, viewsets.ReadOnlyModelViewSet):
    """
    uncached viewset for the `bulbs.content.Tag` model
    """

    model = Tag
    serializer_class = TagSerializer
    paginate_by = 50

    def get_queryset(self):
        queryset = Tag.search_objects.search()
        if "search" in self.request.REQUEST:
            query_string = self.request.REQUEST["search"].lower()
            queryset = queryset.query(Q("match", name=query_string) | Q("match", **{"name.raw": query_string}))
        return queryset

class UserViewSet(UncachedResponse, viewsets.ModelViewSet):
    """
    uncached viewset for the User model -- User can be whatever is defined as the default system
    user model (`auth.User` or any other custom model)
    """

    model = get_user_model()
    serializer_class = UserSerializer
    filter_backends = (filters.SearchFilter, filters.OrderingFilter)
    search_fields = ("first_name", "last_name", "username")
    paginate_by = 20


class LogEntryViewSet(UncachedResponse, viewsets.ModelViewSet):
    """
    uncached viewset for `bulbs.content.LogEntry` model
    """

    model = LogEntry
    serializer_class = LogEntrySerializer

    def get_queryset(self):
        """creates the base queryset object for the serializer

        :return: an instance of `django.db.models.QuerySet`
        """
        qs = LogEntry.objects.all()
        content_id = self.request.QUERY_PARAMS.get("content", None)
        if content_id:
            qs = qs.filter(object_id=content_id)
        return qs

    def create(self, request, *args, **kwargs):
        """
        Grabbing the user from request.user, and the rest of the method
        is the same as ModelViewSet.create().

        :param request: a WSGI request object
        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        :raise: 400
        """
        data = request.DATA.copy()
        data["user"] = request.user.id
        serializer = self.get_serializer(data=data, files=request.FILES)

        if serializer.is_valid():
            self.pre_save(serializer.object)
            self.object = serializer.save(force_insert=True)
            self.post_save(self.object, created=True)
            headers = self.get_success_headers(serializer.data)
            return Response(serializer.data, status=status.HTTP_201_CREATED,
                            headers=headers)

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class AuthorViewSet(UncachedResponse, viewsets.ReadOnlyModelViewSet):
    """
    uncached readonly viewset for the system user model limited to author groups and permissions
    """

    model = get_user_model()
    serializer_class = UserSerializer
    filter_backends = (filters.SearchFilter, filters.OrderingFilter)
    search_fields = ("first_name", "last_name", "username")

    def get_queryset(self):
        """created the base queryset object for the serializer limited to users within the authors
        groups and having `is_staff`

        :return: `django.db.models.QuerySet`
        """
        author_filter = getattr(settings, "BULBS_AUTHOR_FILTER", {"is_staff": True})
        queryset = self.model.objects.filter(**author_filter).distinct()
        return queryset


class FeatureTypeViewSet(UncachedResponse, viewsets.ReadOnlyModelViewSet):
    """
    uncached readonly viewset for the `bulbs.content.FeatureType` model
    """

    model = FeatureType
    serializer_class = FeatureTypeSerializer
    paginate_by = 50

    def get_queryset(self):
        queryset = FeatureType.search_objects.search()
        if "search" in self.request.REQUEST:
            query_string = self.request.REQUEST["search"].lower()
            queryset = queryset.query(Q("match", name=query_string) | Q("match", **{"name.raw": query_string}))
        return queryset


class MeViewSet(UncachedResponse, viewsets.ReadOnlyModelViewSet):
    """
    uncached readonly viewset for users to get information about themselves
    """

    def retrieve(self, request, *args, **kwargs):
        """gets basic information about the user

        :param request: a WSGI request object
        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `rest_framework.response.Response`
        """
        data = UserSerializer().to_representation(request.user)

        # add superuser flag only if user is a superuser, putting it here so users can only
        # tell if they are themselves superusers
        if request.user.is_superuser:
            data['is_superuser'] = True

        # attempt to add a firebase token if we have a firebase secret
        secret = getattr(settings, 'FIREBASE_SECRET', None)
        if secret:
            # use firebase auth to provide auth variables to firebase security api
            firebase_auth_payload = {
                'id': request.user.pk,
                'username': request.user.username,
                'email': request.user.email,
                'is_staff': request.user.is_staff
            }
            data['firebase_token'] = create_token(secret, firebase_auth_payload)

        return Response(data)


class ContentTypeViewSet(viewsets.ViewSet):
    """Searches doctypes of a model."""
    model = Content

    def list(self, request):
        """Search the doctypes for this model."""
        query = request.QUERY_PARAMS.get("search", "")
        results = []
        base = self.model.get_base_class()
        doctypes = indexable_registry.families[base]
        for doctype, klass in doctypes.items():
            name = klass._meta.verbose_name.title()
            if query.lower() in name.lower():
                results.append(dict(
                    name=name,
                    doctype=doctype
                ))
        results.sort(key=lambda x: x["name"])
        return Response(dict(results=results))


class CustomSearchContentViewSet(viewsets.GenericViewSet):
    """This is for searching with a custom search filter."""
    model = Content
    queryset = Content.objects.all()
    serializer_class = ContentSerializer
    paginate_by = 20
    permission_classes = [IsAdminUser, CanEditContent]
    field_map = {
        "feature-type": "feature_type.slug",
        "tag": "tags.slug",
        "content-type": "_type"
    }

    def list(self, request, *args, **kwargs):
        """Filter Content with a custom search.
        {
            "query": SEARCH_QUERY
            "preview": true
        }
        "preview" is optional and, when true, will include
        items that would normally be removed due to "excluded_ids".
        """

        queryset = self.get_filtered_queryset(request.DATA)
        # Switch between paginated or standard style responses
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def create(self, request, *args, **kwargs):
        """HACK: couldn't get POST to the list endpoint without
        messing up POST for the other list_routes so I'm doing this.
        Maybe something to do with the router?
        """
        return self.list(request, *args, **kwargs)

    def get_filtered_queryset(self, params, sort_pinned=True):
        query = params
        is_preview = params.get("preview", True)
        qs = custom_search_model(
            self.model, query, preview=is_preview,
            sort_pinned=sort_pinned, field_map=self.field_map
        )
        return qs.full()

    @list_route(methods=["get", "post"])
    def count(self, request, **kwargs):
        qs = self.get_filtered_queryset(request.DATA, sort_pinned=False)
        return Response(dict(count=qs.count()))

    @list_route(methods=["get", "post"])
    def group_count(self, request, **kwargs):
        params = dict(
            groups=[
                dict(request.DATA)
            ]
        )
        qs = self.get_filtered_queryset(params, sort_pinned=False)
        return Response(dict(count=qs.count()))


# api router for aforementioned/defined viewsets
# note: me view is registered in urls.py
api_v1_router = routers.DefaultRouter()
api_v1_router.register(r"content", ContentViewSet, base_name="content")
api_v1_router.register(r"custom-search-content", CustomSearchContentViewSet, base_name="custom-search-content")
api_v1_router.register(r"content-type", ContentTypeViewSet, base_name="content-type")
api_v1_router.register(r"tag", TagViewSet, base_name="tag")
api_v1_router.register(r"log", LogEntryViewSet, base_name="logentry")
api_v1_router.register(r"author", AuthorViewSet, base_name="author")
api_v1_router.register(r"feature-type", FeatureTypeViewSet, base_name="feature-type")
api_v1_router.register(r"user", UserViewSet, base_name="user")



import json

from dateutil.parser import parse as parse_date

from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.http import Http404
from django.utils import timezone

from rest_framework import viewsets
from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response
from rest_framework.views import APIView

from bulbs.api.permissions import CanPromoteContent
from bulbs.api.mixins import UncachedResponse

from .models import PZone
from .operations import PZoneOperation, InsertOperation, DeleteOperation, ReplaceOperation
from .serializers import PZoneSerializer, InsertOperationSerializer, DeleteOperationSerializer, ReplaceOperationSerializer


class OperationsViewSet(APIView):

    permission_classes = [IsAdminUser, CanPromoteContent]

    def get_serializer_class_by_name(self, type_name):
        try:
            # try to convert type name into a serializer class
            app_label, model_name = type_name.split("_")
            operation_type = ContentType.objects.get_by_natural_key(app_label, model_name)
            operation_model = operation_type.model_class()
            return self.get_serializer_class(operation_model)
        except (ValueError, ContentType.DoesNotExist):
            # invalid type name
            raise ContentType.DoesNotExist("Provided type_name is invalid.")

    def get_serializer_class(self, obj_class):
        serializer_class = None
        if obj_class is InsertOperation:
            serializer_class = InsertOperationSerializer
        elif obj_class is DeleteOperation:
            serializer_class = DeleteOperationSerializer
        elif obj_class is ReplaceOperation:
            serializer_class = ReplaceOperationSerializer
        return serializer_class

    def serialize_operations(self, operations):
        """Serialize a list of operations into JSON."""

        serialized_ops = []
        for operation in operations:
            serializer = self.get_serializer_class(operation.__class__)
            serialized_ops.append(serializer(operation).data)
        return serialized_ops

    def get(self, request, pzone_pk):
        """Get all the operations for a given pzone."""

        # attempt to get given pzone
        try:
            pzone = PZone.objects.get(pk=pzone_pk)
        except PZone.DoesNotExist:
            raise Http404("Cannot find given pzone.")

        # get operations and serialize them
        operations = PZoneOperation.objects.filter(pzone=pzone)

        # return a json response with serialized operations
        return Response(self.serialize_operations(operations), content_type="application/json")

    def post(self, request, pzone_pk):
        """Add a new operation to the given pzone, return json of the new operation."""

        # attempt to get given content list
        pzone = None
        try:
            pzone = PZone.objects.get(pk=pzone_pk)
        except PZone.DoesNotExist:
            raise Http404("Cannot find given pzone.")

        json_obj = []
        http_status = 500

        json_op = json.loads(request.body)
        if not isinstance(json_op, list):
            json_op = [json_op]

        for data in json_op:
            try:
                serializer = self.get_serializer_class_by_name(data["type_name"])
            except ContentType.DoesNotExist as e:
                json_obj = {"errors": [str(e)]}
                http_status = 400
                break

            serialized = serializer(data=data)
            if serialized.is_valid():
                # object is valid, save it
                serialized.save()

                # set response data
                json_obj.append(serialized.data)
                http_status = 200
            else:
                # object is not valid, return errors in a 400 response
                json_obj = serialized.errors
                http_status = 400
                break

        if http_status == 200 and len(json_obj) == 1:
            json_obj = json_obj[0]

        # cache the time in seconds until the next operation occurs
        next_ops = PZoneOperation.objects.filter(when__lte=timezone.now())
        if len(next_ops) > 0:
            # we have at least one operation, ordered soonest first
            next_op = next_ops[0]
            # cache with expiry number of seconds until op should exec
            cache.set('pzone-operation-expiry-' + pzone.name, next_op.when, 60 * 60 * 5)

        return Response(
            json_obj,
            status=http_status,
            content_type="application/json"
        )

    def delete(self, request, pzone_pk, operation_pk):
        """Remove an operation from the given pzone."""

        # note : we're not using the pzone_pk here since it's not actually
        #   necessary for getting an operation by pk, but it sure makes the urls
        #   nicer!

        # attempt to delete operation
        try:
            operation = PZoneOperation.objects.get(pk=operation_pk)
        except PZoneOperation.DoesNotExist:
            raise Http404("Cannot find given operation.")

        # delete operation
        operation.delete()

        # successful delete, return 204
        return Response("", 204)


class PZoneViewSet(UncachedResponse, viewsets.ModelViewSet):
    """Uncached viewset for `bulbs.promotions.PZone` model."""

    queryset = PZone.objects.all()
    serializer_class = PZoneSerializer
    paginate_by = 20
    permission_classes = [IsAdminUser, CanPromoteContent]

    def perform_update(self, serializer):
        """creates a record in the `bulbs.promotion.PZoneHistory`

        :param obj: the instance saved
        :param created: boolean expressing if the object was newly created (`False` if updated)
        """
        instance = serializer.save()
        # create history object
        instance.history.create(data=instance.data)

    def retrieve(self, request, *args, **kwargs):
        """Retrieve pzone as a preview or applied if no preview is provided."""

        when_param = self.request.QUERY_PARAMS.get("preview", None)
        pk = self.kwargs["pk"]

        when = None
        if when_param:
            try:
                when = parse_date(when_param)
            except ValueError:
                # invalid format, set back to None
                when = None

        pzone = None
        if when:
            # we have a date, use it
            pzone = PZone.objects.preview(pk=pk, when=when)
        else:
            # we have no date, just get the pzone
            pzone = PZone.objects.applied(pk=pk)

        # turn content list into json
        return Response(PZoneSerializer(pzone).data, content_type="application/json")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('campaigns', '0002_auto_20150318_1926'),
    ]

    operations = [
        migrations.AlterField(
            model_name='campaignpixel',
            name='pixel_type',
            field=models.IntegerField(default=0, choices=[(0, b'Listing'), (1, b'Detail')]),
        ),
    ]



from django_filters import DateTimeFilter, FilterSet
from rest_framework import routers, viewsets, filters
from rest_framework.permissions import IsAdminUser

from .models import Campaign
from .serializers import CampaignSerializer


class CampaignFilter(FilterSet):
    start_date = DateTimeFilter(lookup_type="lte")
    end_date = DateTimeFilter(lookup_type="gt")

    class Meta(object):
        model = Campaign
        fields = ("start_date", "end_date", )


class CampaignViewSet(viewsets.ModelViewSet):
    queryset = Campaign.objects.all()
    serializer_class = CampaignSerializer
    paginate_by = 10
    # filters
    filter_backends = (filters.DjangoFilterBackend, filters.SearchFilter, filters.OrderingFilter, )
    filter_class = CampaignFilter
    # search
    search_fields = ("campaign_label", "sponsor_name", )
    # ordering
    ordering_fields = ("campaign_label", "sponsor_name", "start_date", "end_date", )
    permission_classes = [IsAdminUser]


api_v1_router = routers.DefaultRouter()
api_v1_router.register(r"campaign", CampaignViewSet, base_name="campaign")



from rest_framework import filters, routers, viewsets
from rest_framework.permissions import IsAdminUser

from .models import SpecialCoverage
from .serializers import SpecialCoverageSerializer


class SpecialCoverageViewSet(viewsets.ModelViewSet):
    queryset = SpecialCoverage.objects.all()
    serializer_class = SpecialCoverageSerializer
    filter_backends = (filters.SearchFilter, filters.OrderingFilter, filters.DjangoFilterBackend,)
    filter_fields = ("active", "promoted")
    search_fields = (
        "name",
        "description",
        "campaign__campaign_label",
        "campaign__sponsor_name"
    )
    ordering_fields = (
        "name",
        "campaign__campaign_label",
        "campaign__sponsor_name"
    )
    paginate_by = 10
    permission_classes = [IsAdminUser]

api_v1_router = routers.DefaultRouter()
api_v1_router.register(
    r"special-coverage",
    SpecialCoverageViewSet,
    base_name="special-coverage")



from collections import OrderedDict

from django.utils import timezone

from bulbs.content.models import Content
from bulbs.content.serializers import UserSerializer

from rest_framework import serializers

from .models import Contribution, ContributorRole


class ContributorRoleSerializer(serializers.ModelSerializer):

    class Meta:
        model = ContributorRole


class ContributionListSerializer(serializers.ListSerializer):

    contributor = UserSerializer()

    def update(self, instance, validated_data):
        # Maps for id->instance and id->data item.
        contribution_mapping = {c.id: c for c in instance}
        data_mapping = {item['id']: item for item in validated_data if "id" in item}

        # Perform creations and updates.
        ret = []
        for data in validated_data:
            contribution = None
            if "id" in data:
                contribution = contribution_mapping.get(data["id"], None)

            if contribution is None:
                ret.append(self.child.create(data))
            else:
                ret.append(self.child.update(contribution, data))

        # Perform deletions.
        for contribution_id, contribution in contribution_mapping.items():
            if contribution_id not in data_mapping:
                contribution.delete()

        return ret


class ContributionSerializer(serializers.ModelSerializer):

    contributor = UserSerializer()
    content = serializers.PrimaryKeyRelatedField(queryset=Content.objects.all())

    class Meta:
        model = Contribution
        list_serializer_class = ContributionListSerializer


class ContributionReportingSerializer(serializers.ModelSerializer):

    user = serializers.SerializerMethodField()
    content = serializers.SerializerMethodField()
    role = serializers.SerializerMethodField()

    class Meta:
        model = Contribution
        fields = ("id", "content", "user", "role", "notes")

    def get_content(self, obj):
        return OrderedDict([
            ("id", obj.content.id),
            ("title", obj.content.title),
            ("url", obj.content.get_absolute_url()),
            ("content_type", obj.content.__class__.__name__),
            ("feature_type", getattr(obj.content.feature_type, "name", None)),
            ("published", timezone.localtime(obj.content.published))
        ])

    def get_user(self, obj):
        return {
            "id": obj.contributor.id,
            "username": obj.contributor.username,
            "full_name": obj.contributor.get_full_name(),
        }

    def get_role(self, obj):
        return obj.role.name


class ContributorRoleField(serializers.Field):
    """This is fucking stupid, but it's basically a field that returns the the
    names of people who have contributed to this content under a certain role."""

    def __init__(self, role, *args, **kwargs):
        super(ContributorRoleField, self).__init__(*args, **kwargs)
        self.role = role
        self.source = "*"

    def to_representation(self, obj):
        qs = Contribution.objects.filter(content=obj, role=self.role).select_related("contributor")
        return ",".join([contribution.contributor.get_full_name() for contribution in qs])


class ContentReportingSerializer(serializers.ModelSerializer):

    content_type = serializers.SerializerMethodField()
    published = serializers.SerializerMethodField()
    feature_type = serializers.SerializerMethodField()
    url = serializers.URLField(source="get_absolute_url")
    authors = serializers.SerializerMethodField()

    class Meta:
        model = Content
        fields = ("id", "title", "url", "content_type", "feature_type", "published", "authors")

    def get_fields(self):

        fields = super(ContentReportingSerializer, self).get_fields()

        self._roles = {}
        for role in ContributorRole.objects.all():
            fields[role.name.lower()] = ContributorRoleField(role)

        return fields

    def get_contributors(self, obj, rolename):
        pass

    def get_content_type(self, obj):
        return obj.__class__.__name__

    def get_feature_type(self, obj):
        return getattr(obj.feature_type, "name", None)

    def get_published(self, obj):
        return timezone.localtime(obj.published)

    def get_authors(self, obj):
        return ",".join([author.get_full_name() for author in obj.authors.all()])



import logging

from django.db import models

from polymorphic import PolymorphicModel

logger = logging.getLogger(__name__)


class PZoneOperation(PolymorphicModel):

    pzone = models.ForeignKey("promotion.PZone", related_name="operations")
    when = models.DateTimeField()
    applied = models.BooleanField(default=False)
    content = models.ForeignKey("content.Content", related_name="+")

    def apply(self, data):
        raise NotImplemented()

    class Meta:
        ordering = ["when", "id"]


class InsertOperation(PZoneOperation):

    index = models.IntegerField(default=0)

    def apply(self, data):

        if self.content.published and self.when >= self.content.published:
            # content has a published date, and that date is before when this

            filtered = filter(lambda content: content["id"] == self.content.pk, data)
            if len(filtered) == 0:
                # content doesn't already exist in list
                data.insert(self.index, {
                    "id": self.content.pk
                })
                data = data[:100]
            else:
                # content already in list, don't insert, log a warning
                logger.warning(
                    "Failed to perform insert operation because content (id: %i) %s is already in %s!",
                    self.content.pk,
                    self.content.title,
                    self.pzone.name)

        else:
            # warn that we failed to perform an operation on this content
            logger.warning(
                "Failed to perform insert operation on unpublished content (id: %i) %s in %s!",
                self.content.pk,
                self.content.title,
                self.pzone.name)

        return data


class ReplaceOperation(PZoneOperation):

    index = models.IntegerField(default=0)

    def apply(self, data):

        if self.content.published and self.when >= self.content.published:
            # content has a published date, and that date is before when this
            #   operation is occurring
            try:
                filtered = filter(lambda content: content["id"] == self.content.pk, data)
                if len(filtered) == 0:
                    # content doesn't already exist in list
                    data[self.index] = {
                        "id": self.content.pk
                    }
                else:
                    # content already in list, don't insert, log a warning
                    logger.warning(
                        "Failed to perform replace operation because content (id: %i) %s is already in %s!",
                        self.content.pk,
                        self.content.title,
                        self.pzone.name)
            except IndexError:
                logger.warning(
                    "Failed to perform replace operation on content (id: %i) %s in %s at index %i!",
                    self.content.pk,
                    self.content.title,
                    self.pzone.name,
                    self.index)
        else:
            # warn that we failed to perform an operation on this content
            logger.warning(
                "Failed to perform replace operation on unpublished content (id: %i) %s in %s!",
                self.content.pk,
                self.content.title,
                self.pzone.name)

        return data


class DeleteOperation(PZoneOperation):
    """Delete a piece of content from the list, assumes only one instance of a
    piece of content is in the list."""

    def apply(self, data):
        for index, item in enumerate(data):
            if item["id"] == self.content.pk:
                del(data[index])
                break
        else:
            logger.warning(
                "Failed to perform delete operation on content (id: %i) %s in %s!",
                self.content.pk,
                self.content.title,
                self.pzone.name)
        return data



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('content', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Contribution',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('notes', models.TextField(null=True, blank=True)),
                ('content', models.ForeignKey(to='content.Content')),
                ('contributor', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ContributorRole',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField(null=True, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='contribution',
            name='role',
            field=models.ForeignKey(to='contributions.ContributorRole'),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings
import djbetty.fields


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Content',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('published', models.DateTimeField(null=True, blank=True)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('title', models.CharField(max_length=512)),
                ('slug', models.SlugField(default=b'', blank=True)),
                ('description', models.TextField(default=b'', max_length=1024, blank=True)),
                ('thumbnail_override', djbetty.fields.ImageField(default=None, null=True, editable=False, blank=True)),
                ('subhead', models.CharField(default=b'', max_length=255, blank=True)),
                ('indexed', models.BooleanField(default=True)),
                ('authors', models.ManyToManyField(to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'permissions': (('publish_own_content', 'Can publish their own content'), ('publish_content', 'Can publish content'), ('promote_content', 'Can promote content')),
            },
        ),
        migrations.CreateModel(
            name='FeatureType',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('slug', models.SlugField(unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='LogEntry',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('action_time', models.DateTimeField(auto_now=True, verbose_name=b'action time')),
                ('object_id', models.TextField(null=True, verbose_name=b'object id', blank=True)),
                ('change_message', models.TextField(verbose_name=b'change message', blank=True)),
                ('content_type', models.ForeignKey(related_name='change_logs', blank=True, to='contenttypes.ContentType', null=True)),
                ('user', models.ForeignKey(related_name='change_logs', blank=True, to=settings.AUTH_USER_MODEL, null=True)),
            ],
            options={
                'ordering': ('-action_time',),
            },
        ),
        migrations.CreateModel(
            name='ObfuscatedUrlInfo',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('create_date', models.DateTimeField()),
                ('expire_date', models.DateTimeField()),
                ('url_uuid', models.CharField(unique=True, max_length=32, editable=False)),
                ('content', models.ForeignKey(to='content.Content')),
            ],
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('slug', models.SlugField(unique=True)),
                ('polymorphic_ctype', models.ForeignKey(related_name='polymorphic_content.tag_set+', editable=False, to='contenttypes.ContentType', null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='content',
            name='feature_type',
            field=models.ForeignKey(blank=True, to='content.FeatureType', null=True),
        ),
        migrations.AddField(
            model_name='content',
            name='polymorphic_ctype',
            field=models.ForeignKey(related_name='polymorphic_content.content_set+', editable=False, to='contenttypes.ContentType', null=True),
        ),
        migrations.AddField(
            model_name='content',
            name='tags',
            field=models.ManyToManyField(to='content.Tag', blank=True),
        ),
    ]



from bulbs.content.models import Content
from django.conf import settings
from django.db import models


# User = get_user_model()
# User = get_model(*settings.AUTH_USER_MODEL.split("."))


class ContributorRole(models.Model):
    name = models.CharField(max_length=255)
    description = models.TextField(null=True, blank=True)


class Contribution(models.Model):
    role = models.ForeignKey(ContributorRole)
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL)
    content = models.ForeignKey(Content)
    notes = models.TextField(null=True, blank=True)



"""API Views and ViewSets"""

import datetime

from django.utils import dateparse, timezone

from rest_framework import viewsets, routers, mixins
from rest_framework.settings import api_settings

from rest_framework_csv.renderers import CSVRenderer

from .models import ContributorRole, Contribution
from .serializers import ContributorRoleSerializer, ContributionReportingSerializer, ContentReportingSerializer
from bulbs.content.models import Content


class ContributorRoleViewSet(viewsets.ModelViewSet):
    queryset = ContributorRole.objects.all()
    serializer_class = ContributorRoleSerializer


class ContentReportingViewSet(viewsets.GenericViewSet, mixins.ListModelMixin):

    renderer_classes = (CSVRenderer, ) + tuple(api_settings.DEFAULT_RENDERER_CLASSES)
    serializer_class = ContentReportingSerializer

    def get_queryset(self):
        now = timezone.now()

        start_date = datetime.datetime(
            year=now.year,
            month=now.month,
            day=1,
            tzinfo=now.tzinfo)
        if "start" in self.request.GET:
            start_date = dateparse.parse_date(self.request.GET["start"])

        end_date = now
        if "end" in self.request.GET:
            end_date = dateparse.parse_date(self.request.GET["end"])

        content = Content.objects.filter(published__range=(start_date, end_date)).prefetch_related("authors", "contribution_set").select_related("feature_type")

        return content


class ReportingViewSet(viewsets.GenericViewSet, mixins.ListModelMixin):

    renderer_classes = (CSVRenderer, ) + tuple(api_settings.DEFAULT_RENDERER_CLASSES)
    serializer_class = ContributionReportingSerializer

    def get_queryset(self):
        now = timezone.now()

        start_date = datetime.datetime(
            year=now.year,
            month=now.month,
            day=1,
            tzinfo=now.tzinfo)
        if "start" in self.request.GET:
            start_date = dateparse.parse_date(self.request.GET["start"])

        end_date = now
        if "end" in self.request.GET:
            end_date = dateparse.parse_date(self.request.GET["end"])

        content = Content.objects.filter(published__range=(start_date, end_date))
        content_ids = content.values_list("pk", flat=True)
        contributions = Contribution.objects.filter(content__in=content_ids)

        ordering = self.request.GET.get("ordering", "content")
        order_options = {
            "content": "content__published",
            "user": "contributor__id"
        }
        return contributions.order_by(order_options[ordering])


api_v1_router = routers.DefaultRouter()
api_v1_router.register(r"role", ContributorRoleViewSet, base_name="contributorrole")
api_v1_router.register(r"reporting", ReportingViewSet, base_name="contributionreporting")
api_v1_router.register(r"contentreporting", ContentReportingViewSet, base_name="contentreporting")



from django.db import models

from djbetty import ImageField

from djes.models import Indexable

from bulbs.content.models import ElasticsearchImageField


class Campaign(Indexable):

    sponsor_name = models.CharField(max_length=255)
    sponsor_logo = ImageField(null=True, blank=True)
    sponsor_url = models.URLField(null=True, blank=True)
    start_date = models.DateTimeField(null=True, blank=True)
    end_date = models.DateTimeField(null=True, blank=True)
    campaign_label = models.CharField(max_length=255)
    impression_goal = models.IntegerField(null=True, blank=True)

    class Mapping:
        sponsor_logo = ElasticsearchImageField()


    @property
    def pixel_dict(self):
        data = {}
        for pixel in self.pixels.all():
            data[pixel.get_pixel_type_display()] = pixel.url
        return data


class CampaignPixel(models.Model):
    """Right now, there are two types of pixels, "Listing" and "Detail". The
    intention here is that the "Listing" pixel is fired anywhere a sponsor's
    logo shows up on a listing page, or a sidebar. The "Detail" pixel is to 
    be fired only when viewing a piece of content connected to that campaign"""
    LISTING = 0
    DETAIL = 1
    PIXEL_TYPES = (
        (LISTING, 'Listing'),
        (DETAIL, 'Detail'),
    )

    campaign = models.ForeignKey(Campaign, related_name='pixels')
    url = models.URLField()
    pixel_type = models.IntegerField(choices=PIXEL_TYPES, default=LISTING)
